# import models into model package
from crtools.models.demerit import Demerit
from crtools.models.formermember import FormerMember
from crtools.models.membercustomrecord import MemberCustomRecord
from crtools.models.memberevent import MemberEvent
from crtools.models.vacation import MemberVacation
from crtools.models.clan import ProcessedClan
from crtools.models.processedmember import ProcessedMember
from crtools.models.war import ProcessedCurrentWar
from crtools.models.warparticipation import WarParticipation



from pyroyale import Clan

from crtools.models import Demerit

class FormerMember():

    def __init__(self, config, historical_member, player_tag, processed_events):
        self.name          = historical_member['name']
        self.tag           = player_tag
        self.blacklist     = False
        self.events        = processed_events
        self.timestamp     = self.events[-1].timestamp
        self.reason        = 'Quit'
        self.notes         = ''

        if player_tag in config['members']['warned']:
        	demerit = config['members']['warned'][player_tag]
        	if type(demerit) is Demerit:
		        self.reason = 'Warned'
        		self.notes = demerit.notes

        if player_tag in config['members']['kicked']:
        	demerit = config['members']['kicked'][player_tag]
        	if type(demerit) is Demerit:
		        self.reason = 'Kicked'
        		self.notes = demerit.notes

        if player_tag in config['members']['blacklist']:
        	self.blacklist = True

        	demerit = config['members']['blacklist'][player_tag]
        	if type(demerit) is Demerit:
		        self.reason = 'Blacklisted'
        		self.notes = demerit.notes



from datetime import datetime #, timedelta
import logging
import math

from pyroyale import Clan

from crtools import leagueinfo

class ProcessedClan():
    def __init__(self, clan, current_war, config):
        self.tag = clan.tag
        self.name = clan.name
        self.badge_id = clan.badge_id
        self.type = clan.type
        self.clan_score = clan.clan_score
        self.required_trophies = clan.required_trophies
        self.donations_per_week = clan.donations_per_week
        self.clan_war_trophies = clan.clan_war_trophies
        self.clan_chest_level = clan.clan_chest_level
        self.clan_chest_max_level = clan.clan_chest_max_level
        self.members = clan.members
        self.location = clan.location
        self.description = clan.description
        self.clan_chest_status = clan.clan_chest_status
        self.clan_chest_points = clan.clan_chest_points

        self.war_league = leagueinfo.get_war_league_from_score(self.clan_war_trophies)
        self.war_league_name = config['strings']['war-league-' + self.war_league]
        self.current_war_state = current_war.state




from datetime import datetime
import logging

from googleapiclient.discovery import build

from crtools.models import Demerit, MemberVacation, MemberCustomRecord

logger = logging.getLogger('gdoc')

DEMERIT_RANGE = 'Demerits!A2:G'
VACATION_RANGE = 'Vacation!A2:E'
CUSTOM_RANGE = 'Custom!A2:D'

def get_member_data_from_sheets(config):
    """ If API key and Sheet ID are provided in config, use the Google Sheets API to
    grab blacklist, no-promote list, and vacation data.

    Template found here: https://docs.google.com/spreadsheets/d/1_8YKfJf-2HVZOgtuosVaGM_50kB8q7YYR3H2d8p0Wzw
    """

    if not config['google_docs']['api_key'] or not config['google_docs']['sheet_id']:
        return config

    sheet_id = config['google_docs']['sheet_id']

    logging.info('Grabbing blacklist/vacation data from Google Sheet {}'.format(sheet_id))

    sheet = get_sheet(config['google_docs']['api_key'])

    (config['members']['blacklist'], config['members']['no_promote'], config['members']['kicked'], config['members']['warned']) = get_demerit_data_from_sheet(sheet, sheet_id, config['members']['blacklist'], config['members']['no_promote'])
    config['members']['vacation'] = get_vacation_data_from_sheet(sheet, sheet_id, config['members']['vacation'])
    config['members']['custom'] = get_custom_data_from_sheet(sheet, sheet_id)

    return config

# no coverage of this function because this is the function we're mocking to get mock
# results for unit tests
def get_sheet(api_key): # pragma: no coverage
    """ Returns authenticated Google Sheet API wrapper using API key """
    try:
        return build('sheets', 'v4', cache_discovery=False, developerKey=api_key).spreadsheets()
    except Exception as e:
        logging.error(e)
        logging.error('Unable to connect to Google Sheets API')

def get_demerit_list(sheet, sheet_id):
    try:
        values = sheet.values() \
                      .get(spreadsheetId=sheet_id, range=DEMERIT_RANGE) \
                      .execute() \
                      .get('values', [])

        demerits = []
        current_tag = ''
        for (member_name, member_tag, action, member_status, reporter, date, notes) in values:
            if member_tag:
                current_tag = member_tag.strip()

            demerits.append(Demerit(tag=current_tag, action=action, status=member_status, date=date, notes=notes))

        return demerits

    except Exception as e:
        logging.error(e)
        logging.error('Unable to get blacklist data from Google Sheets {}'.format(sheet_id))

    return []

def get_vacation_list(sheet, sheet_id):
    try:
        values = sheet.values() \
                      .get(spreadsheetId=sheet_id, range=VACATION_RANGE) \
                      .execute() \
                      .get('values', [])

        now = datetime.utcnow().date()

        vacations = []
        for (member_name, member_tag, start_date, end_date, notes) in values:
            vacation_end = datetime.strptime(end_date, '%m/%d/%Y').date()

            if vacation_end >= now:
                vacations.append(MemberVacation(tag=member_tag, start_date=start_date, end_date=end_date, notes=notes))

        return vacations
    except Exception as e:
        logging.error(e)
        logging.error('Unable to get vacation data from Google Sheets {}'.format(sheet_id))

    return []

def get_custom_record_list(sheet, sheet_id):
    try:
        values = sheet.values() \
                      .get(spreadsheetId=sheet_id, range=CUSTOM_RANGE) \
                      .execute() \
                      .get('values', [])

        now = datetime.utcnow().date()

        vacations = []
        for (member_name, member_tag, custom_role, notes) in values:
            vacations.append(MemberCustomRecord(tag=member_tag, role=custom_role, notes=notes))

        return vacations
    except Exception as e:
        logging.error(e)
        logging.error('Unable to get custom data from Google Sheets {}'.format(sheet_id))

    return []


def get_demerit_data_from_sheet(sheet, sheet_id, blacklist={}, no_promote_list={}, kicked_list={}, warned_list={}):

    demerits = get_demerit_list(sheet, sheet_id)

    if demerits:
        for demerit in demerits:
            if demerit.action == 'kicked':
                kicked_list[demerit.tag] = demerit
            elif demerit.action == 'warning':
                warned_list[demerit.tag] = demerit

            if demerit.status == 'blacklist':
                if demerit.tag in blacklist:
                    demerit.merge(blacklist[demerit.tag])
                blacklist[demerit.tag] = demerit
            elif demerit.status == 'no-promote list':
                no_promote_list[demerit.tag] = demerit

    return (blacklist, no_promote_list, kicked_list, warned_list)

def get_vacation_data_from_sheet(sheet, sheet_id, vacation_list={}):

    vacations = get_vacation_list(sheet, sheet_id)

    if vacations:
        for vacation in vacations:
            vacation_list[vacation.tag] = vacation

    return vacation_list

def get_custom_data_from_sheet(sheet, sheet_id, custom_list={}):

    custom = get_custom_record_list(sheet, sheet_id)

    if custom:
        for record in custom:
            custom_list[record.tag] = record

    return custom_list



from datetime import datetime
from discord_webhook import DiscordEmbed, DiscordWebhook
import logging
import math
from requests.exceptions import ConnectionError

from ._version import __version__

logger = logging.getLogger(__name__)

def trigger_webhooks(config, current_war, member_list):
    if not config['discord']['webhook_default']:
        return False

    return send_war_nag(config, current_war, member_list)

class WarNagConfig:

    abort = False
    nag_header = ''
    webhook_url = ''
    naughty_member_list = ''
    quit_member_list = ''

    def __init__(self, config, current_war, member_list):
        if current_war.state == 'notInWar':
            logger.debug('Not in war; no nagging')
            self.abort = True
            return

        now = datetime.utcnow()
        self.webhook_url = config['discord']['webhook_default']
        if config['discord']['webhook_war_nag']:
            self.webhook_url = config['discord']['webhook_war_nag']

        if current_war.state == 'collectionDay':
            if config['discord']['nag_collection_battle'] == False:
                self.abort = True
                return

            war_end_timestamp = datetime.strptime(current_war.collection_end_time.split('.')[0], '%Y%m%dT%H%M%S')
            nag_threshold = config['discord']['nag_collection_battle_hours_left']
            war_day_label = config['strings']['discord-collection-label']

        elif current_war.state == 'warDay':
            if config['discord']['nag_war_battle'] == False:
                self.abort = True
                return
            war_end_timestamp = datetime.strptime(current_war.war_end_time.split('.')[0], '%Y%m%dT%H%M%S')
            nag_threshold = config['discord']['nag_war_battle_hours_left']
            war_day_label = config['strings']['discord-war-label']

        war_end_time_delta = math.floor((war_end_timestamp - now).seconds / 3600)
        if war_end_time_delta > nag_threshold:
            logger.debug('Nag threshold is {} hours, and there are {} hours left. Not nagging.'.format(nag_threshold, war_end_time_delta))
            self.abort = True
            return

        logger.debug('Compiling list of users to nag')

        self._war_nag_get_naughty_member_list(current_war.participants, member_list)

        self.nag_header = config['strings']['discord-header-war-nag'].format(war_end_time_delta, war_day_label)

    def _war_nag_get_naughty_member_list(self, war_participants, member_list):
        for member in war_participants:
            if member.battles_played < member.number_of_battles:
                member_bullet = '- **{}**\n'.format(escape_markdown(member.name))
                if is_member_in_clan(member_list, member.tag):
                    self.naughty_member_list += member_bullet
                else:
                    self.quit_member_list += member_bullet

def send_war_nag(config, current_war, member_list):

    nag_config = WarNagConfig(config, current_war, member_list)

    if nag_config.abort or (nag_config.naughty_member_list == ''):
        return True

    webhook = DiscordWebhook(url=nag_config.webhook_url)

    # add list of naughty users as embed embed object to webhook
    embed = DiscordEmbed(
        title=nag_config.nag_header,
        description=nag_config.naughty_member_list,
        color = int('0xff5500', 16)
    )

    if nag_config.quit_member_list:
        webhook.add_embed(embed)
        embed = DiscordEmbed(
            title=config['strings']['discord-header-war-quit'].format(),
            description=nag_config.quit_member_list,
            color = int('0x660000', 16)
    )

    embed.set_footer(text='crtools v{}'.format(__version__))
    embed.set_timestamp()

    webhook.add_embed(embed)

    try:
        logger.info('Sending war nag to Discord')
        webhook.execute()
    except ConnectionError as e:
        logger.error('Connection to discord failed. Sending war nag webhook failed.')
        return False

    return True

def escape_markdown(s):
    markdown_escape_map = {'_' : '\\_', '*' : '\\*'}
    for search, replace in markdown_escape_map.items():
        s = s.replace(search, replace)
    return s

def is_member_in_clan(member_list, member_tag):
    for member in member_list:
        if member.tag == member_tag:
            return True
    return False




import codecs
from datetime import datetime, date, timezone, timedelta
from jinja2 import Environment, PackageLoader, StrictUndefined, select_autoescape
import json
import logging
import os
import shutil

from ._version import __version__

logger = logging.getLogger(__name__)

HISTORY_FILE_NAME = 'history.json'

CLAN_LOG_FILENAME = 'clan_logo.png'
FAVICON_FILENAME = 'favicon.ico'

MEMBER_TABLE_CSS_MAPPING = {
    'show_rank'                 : 'rank',
    'show_rank_previous'        : 'rank.previous',
    'show_name'                 : 'name',
    'show_score'                : 'score',
    'show_trophies'             : 'trophies',
    'show_donations'            : 'donations',
    'show_donations_received'   : 'donations-received',
    'show_last_seen'            : 'last-seen',
    'show_days_inactive'        : 'inactivity',
    'show_war_win_rate'         : 'war-win-rate',
    'show_current_war'          : 'war.current',
    'show_warlog'               : 'war.previous'
}

def json_dumper(obj):
    try:
        return obj.to_dict()
    except:
        return obj.__dict__

def write_object_to_file(file_path, obj):
    """ Writes contents of object to file. If object is a string, write it
    directly. Otherwise, convert it to JSON first """

    # open file for UTF-8 output, and write contents of object to file
    with codecs.open(file_path, 'w', 'utf-8') as f:
        if isinstance(obj, str):
            string = obj
        else:
            string = json.dumps(obj, default=json_dumper, indent=4)
        f.write(string)

def get_previous_history(output_dir):
    """ grab history, if it exists, from output paths """

    if not output_dir:
        return None

    history_path = os.path.join(output_dir, HISTORY_FILE_NAME)
    if not os.path.isfile(history_path):
        return None

    with open(history_path, 'r') as myfile:
        return json.loads(myfile.read())

def copy_static_assets(tempdir, clan_logo_path, favicon_path):
    # copy static assets to output path
    shutil.copytree(os.path.join(os.path.dirname(__file__), 'static'), os.path.join(tempdir, 'static'))

    # copy user-provided assets to the output path
    shutil.copyfile(clan_logo_path, os.path.join(tempdir, CLAN_LOG_FILENAME))
    shutil.copyfile(favicon_path, os.path.join(tempdir, FAVICON_FILENAME))


def parse_templates(config, history, tempdir, clan, members, former_members, current_war, recent_wars, suggestions, scoring_rules): # pragma: no coverage
    # Create environment for template parser
    env = Environment(
        loader=PackageLoader('crtools', 'templates'),
        autoescape=select_autoescape(['html', 'xml']),
        undefined=StrictUndefined
    )

    hidden_columns = []
    for key, value in MEMBER_TABLE_CSS_MAPPING.items():
        if config['member_table'][key] != True:
            hidden_columns.append(value)

    dashboard_html = env.get_template('page.html.j2').render(
        version           = __version__,
        config            = config,
        strings           = config['strings'],
        update_date       = datetime.now().strftime('%c'),
        members           = members,
        clan              = clan,
        clan_hero         = config['paths']['description_html_src'],
        current_war       = current_war,
        recent_wars       = recent_wars,
        suggestions       = suggestions,
        scoring_rules     = scoring_rules,
        former_members    = former_members,
        hidden_columns    = hidden_columns
    )

    write_object_to_file(os.path.join(tempdir, 'index.html'), dashboard_html)
    write_object_to_file(os.path.join(tempdir, HISTORY_FILE_NAME), history)

    # If canonical URL is provided, also render the robots.txt and
    # sitemap.xml
    if config['www']['canonical_url'] != False:
        lastmod = datetime.utcnow().replace(tzinfo=timezone.utc).isoformat()
        sitemap_xml = env.get_template('sitemap.xml.j2').render(
                url     = config['www']['canonical_url'],
                lastmod = lastmod
            )
        robots_txt = env.get_template('robots.txt.j2').render(
                canonical_url = config['www']['canonical_url']
            )
        write_object_to_file(os.path.join(tempdir, 'sitemap.xml'), sitemap_xml)
        write_object_to_file(os.path.join(tempdir, 'robots.txt'), robots_txt)

def dump_debug_logs(tempdir, objects_to_dump):
    """ archive outputs of API and other objects for debugging """
    log_path = os.path.join(tempdir, 'log')
    os.makedirs(log_path)
    for name, obj in objects_to_dump.items():
        write_object_to_file(os.path.join(log_path, name+'.json'), obj)

def move_temp_to_output_dir(tempdir, output_dir):
    if os.path.exists(output_dir):
        # remove contents of output directory to cleanup.
        try:
            for file in os.listdir(output_dir):
                file_path = os.path.join(output_dir, file)
                if os.path.isfile(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
        except PermissionError as e:
            logger.error('Permission error: could not delete: \n\t{}'.format(e.filename))
    else:
        # Output directory doesn't exist. Create it.
        logger.debug('Output directory {} doesn\'t exist. Creating it.'.format(output_dir))
        try:
            os.mkdir(output_dir)
        except PermissionError as e: # pragma: no coverage
            logger.error('Permission error: could create output folder: \n\t{}'.format(e.filename))

    # Copy all contents of temp directory to output directory
    # NOTE: not in try/catch block because if the above executed, we already
    # have sufficient privileges to write to the output directory
    for file in os.listdir(tempdir):
        file_path = os.path.join(tempdir, file)
        file_out_path = os.path.join(output_dir, file)
        if os.path.isfile(file_path):
            shutil.copyfile(file_path, file_out_path)
        elif os.path.isdir(file_path):
            shutil.copytree(file_path, file_out_path)



#!/usr/bin/python

"""Hardcoded data about arena and war leagues, and some functions
used to lookup specific data."""

__license__   = 'LGPLv3'
__docformat__ = 'reStructuredText'

import logging
from ._version import __version__

ARENA_LEAGUES = [
    { 'id':  'ultimate-champion', 'trophies': 7000, 'collection_win': { 'bronze': 180, 'silver': 360, 'gold': 630, 'legendary': 990 } },
    { 'id':  'royal-champion',    'trophies': 6600, 'collection_win': { 'bronze': 180, 'silver': 360, 'gold': 630, 'legendary': 990 } },
    { 'id':  'grand-champion',    'trophies': 6300, 'collection_win': { 'bronze': 180, 'silver': 360, 'gold': 630, 'legendary': 990 } },
    { 'id':  'champion',          'trophies': 6000, 'collection_win': { 'bronze': 180, 'silver': 360, 'gold': 630, 'legendary': 990 } },
    { 'id':  'master-3',          'trophies': 5600, 'collection_win': { 'bronze': 170, 'silver': 340, 'gold': 595, 'legendary': 935 } },
    { 'id':  'master-2',          'trophies': 5300, 'collection_win': { 'bronze': 170, 'silver': 340, 'gold': 595, 'legendary': 935 } },
    { 'id':  'master-1',          'trophies': 5000, 'collection_win': { 'bronze': 170, 'silver': 340, 'gold': 595, 'legendary': 935 } },
    { 'id':  'challenger-3',      'trophies': 4600, 'collection_win': { 'bronze': 160, 'silver': 320, 'gold': 560, 'legendary': 880 } },
    { 'id':  'challenger-2',      'trophies': 4300, 'collection_win': { 'bronze': 160, 'silver': 320, 'gold': 560, 'legendary': 880 } },
    { 'id':  'challenger-1',      'trophies': 4000, 'collection_win': { 'bronze': 160, 'silver': 320, 'gold': 560, 'legendary': 880 } },
    { 'id':  'arena-12',          'trophies': 3600, 'collection_win': { 'bronze': 150, 'silver': 300, 'gold': 525, 'legendary': 825 } },
    { 'id':  'arena-11',          'trophies': 3300, 'collection_win': { 'bronze': 140, 'silver': 280, 'gold': 490, 'legendary': 770 } },
    { 'id':  'arena-10',          'trophies': 3000, 'collection_win': { 'bronze': 130, 'silver': 260, 'gold': 455, 'legendary': 715 } },
    { 'id':  'arena-9',           'trophies': 2600, 'collection_win': { 'bronze': 120, 'silver': 240, 'gold': 420, 'legendary': 660 } },
    { 'id':  'arena-8',           'trophies': 2300, 'collection_win': { 'bronze': 110, 'silver': 220, 'gold': 385, 'legendary': 605 } },
    { 'id':  'arena-7',           'trophies': 2000, 'collection_win': { 'bronze': 100, 'silver': 200, 'gold': 350, 'legendary': 550 } },
    { 'id':  'arena-6',           'trophies': 1600, 'collection_win': { 'bronze': 90,  'silver': 180, 'gold': 315, 'legendary': 495 } },
    { 'id':  'arena-5',           'trophies': 1300, 'collection_win': { 'bronze': 80,  'silver': 160, 'gold': 280, 'legendary': 440 } },
    { 'id':  'arena-4',           'trophies': 1000, 'collection_win': { 'bronze': 70,  'silver': 140, 'gold': 245, 'legendary': 385 } },
    { 'id':  'arena-3',           'trophies':  600, 'collection_win': { 'bronze': 60,  'silver': 120, 'gold': 210, 'legendary': 330 } },
    { 'id':  'arena-2',           'trophies':  300, 'collection_win': { 'bronze': 50,  'silver': 100, 'gold': 175, 'legendary': 275 } },
    { 'id':  'arena-1',           'trophies':    0, 'collection_win': { 'bronze': 40,  'silver': 80,  'gold': 140, 'legendary': 220 } }
]

WAR_LEAGUE_LOOKUP = {
    0    : 'bronze',
    600  : 'silver',
    1500 : 'gold',
    3000 : 'legendary'
}

logger = logging.getLogger(__name__)

def get_arena_league_from_trophies(trophies):
    for league in ARENA_LEAGUES:
        if trophies >= league['trophies']:
            return league

def get_collection_win_cards(war_league, league):
    collection_win_lookup = league['collection_win']

    return collection_win_lookup[war_league]

def get_war_league_from_score(clan_score):
    """ Figure out which war league a clan trophy count corresponds to,
    and return war league details. """
    league = 'ERROR'
    for score, lookup_table in WAR_LEAGUE_LOOKUP.items():
        if clan_score >= score:
            league = lookup_table

    return league

def get_war_league_from_war(war, clan_tag):
    """ Figure out which war league a clan was in during a given war. """
    standing = war.standings

    clan_score = 0
    for clan in standing:
        if clan.clan.tag == clan_tag:
            clan_score = clan.clan.clan_score

    return get_war_league_from_score(clan_score)



from argparse import ArgumentParser, RawDescriptionHelpFormatter
import gettext
import logging
import os
import sys

from ._version import __version__
from .crtools import build_dashboard
from .config import load_config_file
from .memberfactory import MemberFactory
from .scorecalc import ScoreCalculator

logger = logging.getLogger(__name__)

def parse_args(argv):
    # parse command line arguments
    parser = ArgumentParser(prog        = 'crtools',
                            description = '''A tool for creating a dashboard for clan participation in
                                             ClashRoyale. See https://developer.clashroyale.com to sign up
                                             for a developer account and create an API key to use with this.''')
    parser.add_argument('--locale',
                        metavar  = 'LOCALE',
                        help     = 'Locale if language other than English is desired'
                        )
    parser.add_argument('--config',
                        metavar  = 'CONFIG-FILE',
                        help     = 'configuration file for this app.')
    parser.add_argument('--api_key',
                        metavar  = 'KEY',
                        help     = 'API key for developer.clashroyale.com')
    parser.add_argument('--clan',
                        metavar  = 'TAG',
                        help    = 'Clan ID from Clash Royale. If it starts with a "#", clan ID must be quoted.')
    parser.add_argument('--out',
                        metavar  = 'PATH',
                        help     = 'Output path for HTML.')
    parser.add_argument('--favicon',
                        metavar  = 'PATH',
                        help     = 'Source path for favicon.ico. If provided, we will copy to the output directory.')
    parser.add_argument('--clan_logo',
                        metavar  = 'PATH',
                        help     = 'Source path for clan logo PNG. Recommended at least 64x64 pizels. If provided, we will copy to the output directory.')
    parser.add_argument('--description',
                        metavar  = 'PATH',
                        help     = 'Source path snippet of HTML to replace the clan description. Should not be a complete HTML document. Sample here: https://github.com/AaronTraas/crtools-agrassar-assets/blob/master/description.html\n\nIf provided, we will copy to the output directory.')
    parser.add_argument('--canonical_url',
                        metavar  = 'URL',
                        help     = 'Canonical URL for this site. Used for setting the rel=\'canonical\' link in the web site, as well as generating the robots.txt and sitemap.xml')
    parser.add_argument('--debug',
                        action   = 'store_true',
                        help     = 'Turns on debug mode')
    parser.add_argument('--version',
                        action   = 'store_true',
                        help     = 'List the version of crtools.')

    return parser.parse_args(argv)

def get_config_from_args(args, check_config_file=True, config_default='~/.crtools'):
    if args.version:
        print(__version__)
        exit(0)

    if args.config:
        config_file_name = args.config
        os.path.expanduser(config_file_name)
        if os.path.isfile(config_file_name) == False:
            logger.error('Config file specified {} not found'.format(config_file_name) )
            exit(1)
    else:
        config_file_name = os.path.expanduser(config_default)

    locale = None
    if args.locale:
        locale = args.locale

    config = load_config_file(config_file_name, check_config_file, locale)

    # grab API key and clan ID from arguments if applicable
    if args.api_key:
        config['api']['api_key'] = args.api_key
    if args.clan:
        config['api']['clan_id'] = args.clan
    if args.out:
        config['paths']['out'] = args.out
    if args.favicon:
        config['paths']['favicon'] = args.favicon
    if args.clan_logo:
        config['paths']['clan_logo'] = args.clan_logo
    if args.description:
        config['paths']['description_html'] = args.description
    if args.canonical_url:
        config['www']['canonical_url'] = args.canonical_url
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
        config['crtools']['debug'] = True

    return config

def main(): # pragma: no cover
    args = parse_args(sys.argv[1:])

    config = get_config_from_args(args)

    # Build the dashboard
    build_dashboard(config)



class Demerit:
    def __init__(self, tag, status, action=None, date=0, notes=""):
        self.tag = tag.strip()
        self.action = action
        self.status = status
        self.date = date
        self.notes = notes

    def __str__(self):
        return str(self.__dict__)

    def merge(self, other_demerit):
        self.notes = '{}<br>- {}'.format(other_demerit.notes, self.notes)
        if not self.notes.startswith('<br>- '):
            self.notes = '<br>- {}'.format(self.notes)



from datetime import datetime
from html import escape

import logging

from pyroyale import ClanMember

from crtools import leagueinfo
from crtools.scorecalc import ScoreCalculator
from crtools.models.warparticipation import WarParticipation

logger = logging.getLogger(__name__)

class ProcessedMember():
    def __init__(self, member, member_history=[]):
        self.tag = member.tag
        self.name = escape(member.name)
        self.exp_level = member.exp_level
        self.trophies = member.trophies
        self.role = member.role
        self.last_seen = member.last_seen
        self.clan_rank = member.clan_rank
        self.previous_clan_rank = member.previous_clan_rank
        self.donations = member.donations
        self.donations_received = member.donations_received
        self.clan_chest_points = member.clan_chest_points

        self.arena_league = leagueinfo.get_arena_league_from_trophies(self.trophies)

        self.score = 'int'
        self.vacation = False
        self.safe = False
        self.blacklist = False
        self.no_promote = False




from datetime import datetime
import logging

from crtools.models import Demerit, ProcessedMember, WarParticipation
from crtools import history
from crtools.scorecalc import ScoreCalculator

logger = logging.getLogger(__name__)

class MemberFactory:
    def __init__(self, config, clan, member_history, current_war, warlog=[], days_from_donation_reset=0, now=datetime.utcnow()):
        self.config = config
        self.clan = clan
        self.current_war = current_war
        self.warlog = warlog
        self.member_history = member_history
        self.days_from_donation_reset = days_from_donation_reset
        self.now = now

    def get_processed_member(self, member):
        processed_member = ProcessedMember(member)

        self.enrich_member_with_history(processed_member, self.member_history['members'][processed_member.tag])
        self.calc_special_status(processed_member)
        self.calc_derived_member_stats(processed_member)

        return processed_member

    def enrich_member_with_history(self, member, historical_member):
        days_from_donation_reset = self.days_from_donation_reset

        member.join_date = historical_member['join_date']
        member.last_activity_date = historical_member['last_activity_date']
        member.last_donation_date = historical_member['last_donation_date']
        member.donations_last_week = historical_member['donations_last_week']
        member.days_inactive = (self.now - datetime.fromtimestamp(member.last_activity_date)).days
        member.days_inactive = member.days_inactive if member.days_inactive >= 0 else 0

        if member.last_seen:
            last_seen = datetime.strptime(member.last_seen.split('.')[0], '%Y%m%dT%H%M%S')
        else:
            last_seen = datetime.utcnow()

        member.last_seen_formatted = last_seen.strftime('%c')


        last_seen_delta = self.now - last_seen
        member.last_seen_delta = ''
        if last_seen_delta.days >= 1:
            member.last_seen_delta = '{} {}, '.format(last_seen_delta.days, self.config['strings']['labelDays'])
        hours = round(last_seen_delta.seconds/3600)
        if hours < 1:
            member.last_seen_delta += '{} {}'.format(round(last_seen_delta.seconds/60), self.config['strings']['labelMinutes'])
        else:
            member.last_seen_delta += '{} {}'.format(hours, self.config['strings']['labelHours'])

        if member.join_date == 0:
            member.join_date_label = self.config['strings']['labelBeforeHistory']
        else:
            member.join_date_label = datetime.fromtimestamp(member.join_date).strftime('%x')
        member.activity_date_label = datetime.fromtimestamp(member.last_activity_date).strftime('%x')

        join_datetime = datetime.fromtimestamp(member.join_date)
        member.days_from_join = (self.now - join_datetime).days
        if member.days_from_join <= 10:
            member.new = True
            logger.debug('New member {}'.format(member.name))
        else:
            member.new = False

        if days_from_donation_reset > member.days_from_join:
            days_from_donation_reset = member.days_from_join

        if member.days_inactive > 7:
            member.donations_last_week = 0

        member.total_donations = member.donations + member.donations_last_week
        if member.days_from_join > days_from_donation_reset + 7:
            days_from_donation_reset += 7
        else:
            days_from_donation_reset = member.days_from_join

        if(days_from_donation_reset > 0):
            member.donations_daily = round(member.total_donations / days_from_donation_reset)
        else:
            member.donations_daily = member.total_donations

        member.events = history.process_member_events(self.config, historical_member['events'])

    def calc_special_status(self, member):
        # get special statuses.
        # vacation = member is on vacation. Don't reccomend demote or kick, dont show score
        # safe = member marked as safe. Don't reccomend demote or kick
        # blacklist = member on blacklist. Recommend kick immediately.
        member.vacation = False
        if member.tag in self.config['members']['vacation']:
            member.vacation = self.config['members']['vacation'][member.tag]

        member.safe = member.tag in self.config['members']['safe']
        member.no_promote = member.tag in self.config['members']['no_promote']
        member.blacklist = member.tag in self.config['members']['blacklist']
        member.notes = ''

        if member.tag in self.config['members']['no_promote']:
            demerit = self.config['members']['no_promote'][member.tag]
            if type(demerit) is Demerit:
                member.notes = demerit.notes

        if member.tag in self.config['members']['blacklist']:
            demerit = self.config['members']['blacklist'][member.tag]
            if type(demerit) is Demerit:
                member.notes = demerit.notes

        # Automatically add inactive 'safe' members to vacation
        if member.safe and (member.days_inactive >= self.config['activity']['threshold_warn']):
            member.vacation = True

    def calc_derived_member_stats(self, member):
        member.current_war = WarParticipation(self.config, member, self.current_war)
        member.warlog = []
        for war in self.warlog.items:
            member.warlog.append(WarParticipation(self.config, member, war))

        score_calc = ScoreCalculator(self.config)

        member.donation_score = score_calc.get_member_donations_score(member)

        # calculate score based on war participation
        member.war_score = 0
        for war in member.warlog:
            member.war_score += war.score

        # get member score
        member.score = member.war_score + member.donation_score

        # members on the safe list can't have a score below zero
        if member.safe and member.score < 0:
            member.score = 0

        # calculate the number of daily donations, and the donation status
        # based on threshold set in config
        member.donation_status = self.calc_donation_status(member.donation_score, member.donations_daily, self.days_from_donation_reset)

        member.status = self.calc_member_status(member.score, member.no_promote)

        member.activity_status = self.calc_activity_status(member.days_inactive)

        member.role_label = self.get_role_label(member.tag, member.role, member.days_inactive, member.activity_status, member.vacation, member.blacklist, member.no_promote)

        if member.trophies >= self.clan.required_trophies:
            member.trophies_status = 'normal'
        else:
            member.trophies_status = 'ok'

        member.arena_league_label = self.config['strings']['league-' + member.arena_league['id']]

        # Figure out whether member is on the leadership team by role
        member.leadership = member.role == 'leader' or member.role == 'coLeader'

        self.calc_recent_war_stats(member)

    def get_role_label(self, member_tag, member_role, days_inactive, activity_status, vacation, blacklisted, no_promote):
        """ Format roles in sane way """

        if member_tag in self.config['members']['custom']:
            return self.config['members']['custom'][member_tag].role

        if blacklisted:
            return self.config['strings']['roleBlacklisted']

        if vacation:
            if vacation.end_date == 0:
                return self.config['strings']['roleVacation']
            else:
                return self.config['strings']['roleVacationUntil'].format(vacation.end_date)

        if activity_status in ['bad', 'ok']:
            return self.config['strings']['roleInactive'].format(days=days_inactive)

        if no_promote:
            return self.config['strings']['roleNoPromote']

        return {
            'leader'   : self.config['strings']['roleLeader'],
            'coLeader' : self.config['strings']['roleCoLeader'],
            'elder'    : self.config['strings']['roleElder'],
            'member'   : self.config['strings']['roleMember'],
        }[member_role]

    def calc_recent_war_stats(self, member):
        war_wins = 0
        war_battles = 0
        collection_wins = 0
        collection_cards = 0
        for war in member.warlog:
            if hasattr(war, 'wins'):
                war_wins += war.wins
                war_battles += war.number_of_battles
            if hasattr(war, 'collection_battle_wins'):
                collection_wins += war.collection_battle_wins
            if hasattr(war, 'collection_win_cards'):
                collection_cards += war.collection_win_cards

        if war_battles > 0:
            member.war_win_rate = round((war_wins/war_battles) * 100)
        else:
            member.war_win_rate = 0

        member.war_battles = war_battles
        member.war_collection_win_rate = round(((collection_wins / 10) / 3) * 100)
        member.war_collection_cards_average = round(collection_cards / 10)
        member.war_score_average = round(member.war_score / 10)

    def calc_donation_status(self, donation_score, donations_daily, days_from_donation_reset):
        """ calculate the number of daily donations, and the donation status
        based on threshold set in config """
        if donation_score >= self.config['score']['max_donations_bonus']:
            return 'good'

        if days_from_donation_reset >= 1:
            if donations_daily == 0:
                return 'bad'

            if donations_daily < self.config['score']['min_donations_daily']:
                return 'ok'

        return 'normal'

    def calc_member_status(self, member_score, no_promote):
        # either 'good', 'ok', 'bad', or 'normal'
        if member_score < 0:
            return 'bad'

        if member_score >= self.config['score']['threshold_promote'] and not no_promote:
            return 'good'

        if member_score < self.config['score']['threshold_warn']:
            return 'ok'

        return 'normal'

    def calc_activity_status(self, days_inactive):
        if days_inactive <= 0:
            return 'good'

        if days_inactive <= 2:
            return 'na'

        if days_inactive >= self.config['activity']['threshold_kick']:
            return 'bad'

        if days_inactive >= self.config['activity']['threshold_warn']:
            return 'ok'

        return 'normal'





from datetime import datetime, timedelta

from crtools import leagueinfo
from crtools.scorecalc import ScoreCalculator

def _get_war_date(war):
    """ returns the datetime this war was created. If it's an ongoing
    war, calculate based on the dates given when the war started.
    If it's a previous war fromt he warlog, we retrieve the creation
    date. What's returned is a timestamp. """
    war_date_raw = 0

    if hasattr(war, 'state') :
        if war.state == 'warDay':
            war_date_raw = datetime.strptime(war.war_end_time.split('.')[0], '%Y%m%dT%H%M%S')
            war_date_raw -= timedelta(days=2)
        elif war.state == 'collectionDay':
            war_date_raw = datetime.strptime(war.collection_end_time.split('.')[0], '%Y%m%dT%H%M%S')
            war_date_raw -= timedelta(days=1)
    else:
        war_date_raw = datetime.strptime(war.created_date.split('.')[0], '%Y%m%dT%H%M%S')
        war_date_raw -= timedelta(days=1)

    return datetime.timestamp(war_date_raw)


def _get_member_war_status_class(collection_day_battles, war_day_battles, war_date, join_date, current_war=False, war_day=False):
    """ returns CSS class(es) for a war log entry for a given member """
    if war_date < join_date:
        return 'not-in-clan'

    status = 'normal'
    if current_war:
        if collection_day_battles < 3:
            status = 'ok'
        elif war_day and war_day_battles > 0:
            status = 'good'

        if war_day == False or war_day_battles == 0:
            status += ' incomplete'
    else:
        if collection_day_battles == 0:
            status = 'na'
        elif war_day_battles == 0:
            status = 'bad'
        elif collection_day_battles < 3:
            status = 'ok'
        else:
            status = 'good'
    return status

class WarParticipation():

    def __init__(self, config, member, war):
        self.in_war = False
        self.status = 'na'
        self.score = 0
        self.battles_played = False
        self.cards_earned = 0
        self.collection_win_cards = 0
        self.collection_day_battles_played = 0
        self.collection_battle_wins = 0
        self.collection_battle_losses = 0
        self.wins = 0
        self.war_league = 'na'
        self.number_of_battles = 0

        # Bail early if this is for the current war, and there is no
        # current war
        if hasattr(war, 'state') and war.state == 'notInWar':
            return

        member_tag = member.tag
        war_date = _get_war_date(war)
        join_date = member.join_date if hasattr(member, 'join_date') else 0

        if hasattr(war, 'state') :
            self.score = 0
        elif war_date < join_date:
            # member is not in this war
            self.status = 'not-in-clan'
            return
        else:
            # member is not in this war
            self.score = ScoreCalculator(config).get_war_score(self)

        for participant in war.participants:
            if participant.tag == member_tag:
                self.in_war = True
                self.cards_earned = participant.cards_earned
                self.battles_played = participant.battles_played
                self.collection_day_battles_played = participant.collection_day_battles_played
                self.wins = participant.wins
                self.number_of_battles = participant.number_of_battles

                if hasattr(war, 'state'):
                    self.status = _get_member_war_status_class(self.collection_day_battles_played, self.battles_played, war_date, join_date, True, war.state=='warDay')
                    return;

                self.status = _get_member_war_status_class(self.collection_day_battles_played, self.battles_played, war_date, join_date)

                self.war_league = leagueinfo.get_war_league_from_war(war, config['api']['clan_id'])
                self.collection_win_cards = leagueinfo.get_collection_win_cards(self.war_league, member.arena_league)

                self.collection_battle_wins = round(self.cards_earned / self.collection_win_cards)
                self.collection_battle_losses = self.collection_day_battles_played - self.collection_battle_wins
                self.score = ScoreCalculator(config).get_war_score(self)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
import re

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# single-sourcing the version
with open(path.join(here, 'crtools/_version.py')) as f:
    exec(f.read())

setup(
    name='crtools',

    # Version single-sourced from crtools/_version.py
    version=__version__,

    description='Python tools for creating a clan management dashboard for Clash Royale',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/AaronTraas/Clash-Royale-Clan-Tools',

    # Author details
    author='Aaron Traas',
    author_email='aaron@traas.org',

    # Choose your license
    license='LGPLv3+',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python :: 3',
    ],

    keywords='ClashRoyale',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    setup_requires=['babel'],
    install_requires=['jinja2','configparser','pyroyale>=1.0.3', 'requests', 'discord-webhook', 'google-api-python-client'],

    include_package_data=True,

    entry_points={
        'console_scripts': [
            'crtools=crtools:main',
        ],
    },
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/AaronTraas/Clash-Royale-Clan-Tools/issues',
        'Source': 'https://github.com/AaronTraas/Clash-Royale-Clan-Tools',
    },
)



from datetime import datetime
import logging
import math

from pyroyale import WarCurrent

class ProcessedCurrentWar:

    cards = 0
    state_label = ''
    collection_end_time_label = ''
    end_label = ''

    def __init__(self, current_war, config):
        self.state = current_war.state
        self.war_end_time = current_war.war_end_time
        self.collection_end_time = current_war.collection_end_time
        self.clan = current_war.clan
        self.participants = current_war.participants
        self.clans = current_war.clans

        self.cards = 0
        self.state_label = ''
        self.collection_end_time_label = ''
        self.end_label = ''

        self.process(config)

    def process(self, config):
        if self.state == 'notInWar':
            self.state_label = config['strings']['LabelNotInWar']
            return

        self.cards = 0;
        for member in self.participants:
            self.cards += member.cards_earned

        now = datetime.utcnow()
        if self.state == 'collectionDay':
            self.state_label = config['strings']['labelStateCollectionDay']

            collection_end_time = datetime.strptime(self.collection_end_time.split('.')[0], '%Y%m%dT%H%M%S')
            collection_end_time_delta = math.floor((collection_end_time - now).seconds / 3600)
            self.collection_end_time_label = config['strings']['labelCollectionEndTime'].format(collection_end_time_delta)
            self.end_label = config['strings']['labelEndTime'].format(collection_end_time_delta)
        else:
            self.state_label = config['strings']['labelStateWarDay']

            end_time = datetime.strptime(self.war_end_time.split('.')[0], '%Y%m%dT%H%M%S')
            end_time_delta = math.floor((end_time - now).seconds / 3600)
            self.collection_end_time_label = config['strings']['labelCollectionComplete']
            self.end_label = config['strings']['labelCollectionEndTime'].format(end_time_delta)

            # calculate battles remaining for each clan
            for clan in self.clans:
                clan.battles_remaining = clan.participants - clan.battles_played
                if clan.battles_remaining < 0:
                    clan.battles_remaining = 0; # pragma: no coverage

            # sort clans by who's winning
            self.clans = sorted(self.clans, key=lambda k: (k.wins, k.crowns), reverse=True)




from datetime import datetime

class MemberEvent():
    def __init__(self, config, event_dict):
        self.date      = datetime.fromtimestamp(event_dict['date']).strftime('%x')
        self.timestamp = event_dict['date']
        self.event     = event_dict['event']
        self.message   = {
            'join'        : config['strings']['memberEventJoinedClan'],
            'role change' : config['strings']['memberEventRoleChange'].format(event_dict['role']),
            'quit'        : config['strings']['memberEventExitClan']
        }[self.event]





class ScoreCalculator:
    def __init__(self, config):
        self.config = config

    def get_member_donations_score(self, member):
        """ Calculate the score for a given member's daily donations. """

        donation_score = member.donations_daily - self.config['score']['min_donations_daily']

        donation_score = donation_score if donation_score <= self.config['score']['max_donations_bonus'] else self.config['score']['max_donations_bonus']

        if member.total_donations == 0:
            donation_score += self.config['score']['donations_zero']

        if member.days_from_join <= self.config['score']['new_member_grace_period_days'] and donation_score <= 0:
            donation_score = 0

        return donation_score

    def get_war_score(self, war):
        """ Tally the score for a given war """

        if war.status == 'not-in-clan':
            return 0;

        if hasattr(war, 'in_war') and not war.in_war:
            return self.config['score']['war_non_participation']

        war_score = 0
        war_score += war.collection_battle_wins * self.config['score']['collect_battle_won']
        war_score += war.collection_battle_losses * self.config['score']['collect_battle_lost']
        war_score += war.collection_day_battles_played * self.config['score']['collect_battle_played']
        war_score += (3-war.collection_day_battles_played) * self.config['score']['collect_battle_incomplete']

        if war.battles_played < 1:
            war_score += self.config['score']['war_battle_incomplete']
            return war_score

        war_score += self.config['score']['war_battle_played'] if war.battles_played > 0 else 0
        war_score += war.wins * self.config['score']['war_battle_won']
        war_score += (war.battles_played - war.wins) * self.config['score']['war_battle_lost']

        return war_score




#!/usr/bin/python
"""Tools for creating a clan maagement dashboard for Clash Royale."""

__license__   = 'LGPLv3'
__docformat__ = 'reStructuredText'

from datetime import datetime #, date, timezone, timedelta
import logging
import os
import shutil
import tempfile

import pyroyale
import json

from ._version import __version__
from crtools import history
from crtools import fankit
from crtools import io
from crtools import discord
from crtools.memberfactory import MemberFactory
from crtools.models import FormerMember, ProcessedClan, ProcessedCurrentWar

MAX_CLAN_SIZE = 50

logger = logging.getLogger(__name__)

def get_suggestions(config, processed_members, required_trophies):
    """ Returns list of suggestions for the clan leadership to perform.
    Suggestions are to kick, demote, or promote. Suggestions are based on
    user score, and various thresholds in configuration. """

    # sort members by score, and preserve trophy order if relevant
    members_by_score = sorted(processed_members, key=lambda m: (m.score, m.trophies))

    logger.debug("min_clan_size: {}".format(config['score']['min_clan_size']))
    logger.debug("# members: {}".format(len(members_by_score)))

    suggestions = []
    for index, member in enumerate(members_by_score):
        if member.blacklist:
            suggestion = config['strings']['suggestionKickBlacklist'].format(name=member.name)
            logger.debug(suggestion)
            suggestions.append(suggestion)
            continue

        # if member on the 'safe' or 'vacation' list, don't make
        # recommendations to kick or demote
        if not (member.safe or member.vacation) and member.current_war.status == 'na':
            # suggest kick if inactive for the set threshold
            if member.days_inactive >= config['activity']['threshold_kick']:
                suggestion = config['strings']['suggestionKickInactivity'].format(name=member.name, days=member.days_inactive)
                logger.debug(suggestion)
                suggestions.append(suggestion)
            # if members have a score below zero, we recommend to kick or
            # demote them.
            # if we're above the minimum clan size, recommend kicking
            # poorly participating member.
            elif member.score < config['score']['threshold_kick'] and index <= len(members_by_score) - config['score']['min_clan_size']:
                suggestion = config['strings']['suggestionKickScore'].format(name=member.name, score=member.score)
                logger.debug(suggestion)
                suggestions.append(suggestion)
            # If we aren't recommending kicking someone, and their role is
            # > member, recoomend demotion
            elif member.role != 'member' and member.score < config['score']['threshold_demote']:
                suggestions.append(config['strings']['suggestionDemoteScore'].format(name=member.name, score=member.score))

        # if user is above the threshold, and has not been promoted to
        # Elder or higher, recommend promotion.
        if not member.no_promote and not member.blacklist and (member.score >= config['score']['threshold_promote']) and (member.role == 'member') and (member.trophies >= required_trophies) and (member.days_from_join > config['activity']['min_days_to_promote']):
            suggestions.append(config['strings']['suggestionPromoteScore'].format(name=member.name, score=member.score))

    # If there are no other suggestions, give some sort of message
    if len(suggestions) == 0:
        if len(members_by_score) < MAX_CLAN_SIZE:
            suggestions.append(config['strings']['suggestionRecruit'])
        else:
            suggestions.append(config['strings']['suggestionNone'])

    return suggestions

def get_scoring_rules(config):
    """ Get list of scoring rules to display on the site """

    def get_score_rule_status(score):
        if score > 0:
            return 'good'
        elif score < 0:
            return 'bad'
        else:
            return 'normal'

    rules = [
        {'name': config['strings']['ruleParticipate'],         'yes': config['score']['war_participation'],        'no': config['score']['war_non_participation'] },
        {'name': config['strings']['ruleCollectionComplete'], 'yes': config['score']['collect_battle_played'],    'no': config['score']['collect_battle_incomplete']},
        {'name': config['strings']['ruleCollectionWin'],      'yes': config['score']['collect_battle_won'],       'no': config['score']['collect_battle_lost']},
        {'name': config['strings']['ruleWarDayComplete'],     'yes': config['score']['war_battle_played'],        'no': config['score']['war_battle_incomplete']},
        {'name': config['strings']['ruleWarDayWin'],          'yes': config['score']['war_battle_won'],           'no': config['score']['war_battle_lost']}
    ]

    for rule in rules:
        rule['yes_status'] = get_score_rule_status(rule['yes'])
        rule['no_status'] = get_score_rule_status(rule['no'])

    return rules

def process_members(config, clan, warlog, current_war, member_history):
    """ Process member list, adding calculated meta-data for rendering of
    status in the clan member table. """

    # calculate the number of days since the donation last sunday, for
    # donation tracking purposes:
    now = datetime.utcnow()
    days_from_donation_reset = now.isoweekday()
    if days_from_donation_reset > 7 or days_from_donation_reset <= 0:
        days_from_donation_reset = 1

    # process members with results from the API
    factory = MemberFactory(
        config=config,
        clan=clan,
        current_war=current_war,
        warlog=warlog,
        member_history=member_history,
        days_from_donation_reset=days_from_donation_reset,
        now=now)
    members_processed = []
    for member_src in clan.member_list:
        members_processed.append(factory.get_processed_member(member_src))

    return members_processed

def process_absent_members(config, historical_members):
    absent_members = []

    for tag, member in historical_members.items():
        if member['status'] == 'absent':
            absent_members.append(FormerMember(
                config=config,
                historical_member=member,
                player_tag=tag,
                processed_events=history.process_member_events(config, member['events'])
            ))

    return sorted(absent_members, key=lambda k: k.timestamp, reverse=True)

def process_recent_wars(config, warlog):
    wars = []
    for war in warlog.items:
        clan = None
        for rank, war_clan in enumerate(war.standings):
            if war_clan.clan.tag == config['api']['clan_id']:
                clan = war_clan
                clan.rank = rank+1
                date = datetime.strptime(war.created_date.split('.')[0], '%Y%m%dT%H%M%S')
                clan.date = config['strings']['labelWarDate'].format(month=date.month, day=date.day)
                wars.append(clan)

    return wars

def get_data_from_api(config): # pragma: no coverage
    # get API instance
    configuration = pyroyale.Configuration()
    configuration.api_key['authorization'] = config['api']['api_key']
    if config['api']['proxy']:
        configuration.proxy = config['api']['proxy']
    if config['api']['proxy_headers']:
        configuration.proxy_headers = config['api']['proxy_headers']

    api = pyroyale.ClansApi(pyroyale.ApiClient(configuration))

    try:
        # Get clan data and war log from API.
        clan = api.get_clan(config['api']['clan_id'])
        warlog = api.get_clan_war_log(config['api']['clan_id'])
        current_war = api.get_current_war(config['api']['clan_id'])

        print('- clan: {} ({})'.format(clan.name, clan.tag))

        return (clan, warlog, current_war)
    except pyroyale.ApiException as e:
        if e.body:
            body = json.loads(e.body)
            if body['reason'] == 'accessDenied':
                logger.error('developer.clashroyale.com claims that your API key is invalid. Please make sure you are setting up crtools with a valid key.')
            elif body['reason'] == 'accessDenied.invalidIp':
                logger.error('developer.clashroyale.com says: {}'.format(body['message']))
            else:
                logger.error('error: {}'.format(body))
        else:
            logger.error('error: {}'.format(e))
    except pyroyale.OpenApiException as e:
        logger.error('error: {}'.format(e))

    # If we've gotten here, something has gone wrong. We need to abort the application.
    exit(0)

# NOTE: we're not testing this function because this is where we're
# isolating all of the I/O for the application here. The real "work"
# here is done in all of the calls to functions in this file, or in the
# ClashRoyaleAPI class, both of which are fully covered. (or soon will
# be)
#
# Similarly, we've tagged this function, and this function alone, to
# be ignored by static analysis. I don't want to spread out all of
# the I/O and there's no way to make the exception handling anything
# other than a mess that will trigger teh cognitive complexity
# warnings.
def build_dashboard(config): # pragma: no coverage
    """Compile and render clan dashboard."""

    print('- requesting info for clan id: {}'.format(config['api']['clan_id']))

    clan, warlog, current_war = get_data_from_api(config)

    # Create temporary directory. All file writes, until the very end,
    # will happen in this directory, so that no matter what we do, it
    # won't hose existing stuff.
    tempdir = tempfile.mkdtemp(config['paths']['temp_dir_name'])

    # Putting everything in a `try`...`finally` to ensure `tempdir` is removed
    # when we're done. We don't want to pollute the user's disk.
    try:
        output_path = os.path.expanduser(config['paths']['out'])

        # process data from API
        current_war_processed = ProcessedCurrentWar(current_war, config)
        clan_processed = ProcessedClan(clan, current_war_processed, config)

        member_history = history.get_member_history(clan.member_list, io.get_previous_history(output_path), current_war_processed)

        members_processed = process_members(config, clan, warlog, current_war_processed, member_history)
        recent_wars = process_recent_wars(config, warlog)
        former_members = process_absent_members(config, member_history['members'])

        io.parse_templates(
            config,
            member_history,
            tempdir,
            clan_processed,
            members_processed,
            former_members,
            current_war_processed,
            recent_wars,
            get_suggestions(config, members_processed, clan_processed.required_trophies),
            get_scoring_rules(config)
        )

        if(config['crtools']['debug'] == True):
            # archive outputs of API for debugging
            io.dump_debug_logs(
                tempdir,
                {
                    'clan'                  : clan.to_dict(),
                    'warlog'                : warlog.to_dict(),
                    'current_war'           : current_war.to_dict(),
                    'clan-processed'        : clan_processed,
                    'members-processed'     : members_processed,
                    'current_war-processed' : current_war_processed,
                    'recentwars-processed'  : list(map(lambda war: war.to_dict(), recent_wars))
                }
            )

        # if fankit is previously downloaded, it will copy fankit. Otherwise,
        # if fankit is enabled, it will download it.
        fankit.get_fankit(tempdir, output_path, config['paths']['use_fankit'])

        io.copy_static_assets(tempdir, config['paths']['clan_logo'], config['paths']['favicon'])

        io.move_temp_to_output_dir(tempdir, output_path)

        discord.trigger_webhooks(config, current_war, members_processed)

    except Exception as e:
        logger.error('error: {}'.format(e))

    finally:
        # Ensure that temporary directory gets deleted no matter what
        shutil.rmtree(tempdir)
